package com.streambase.sb.adapter.component.http.client;
/*
* Copyright � 2019. TIBCO Software Inc.
* This file is subject to the license terms contained
* in the license file that is distributed with this file.
*/


import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.slf4j.Logger;

import com.streambase.sb.*;
import com.streambase.sb.adapter.*;
import com.streambase.sb.adapter.common.AdapterUtil;
import com.streambase.sb.adapter.common.status.StatusPort;
import com.streambase.sb.operator.*;
import com.streambase.sb.util.Util;

import org.glassfish.jersey.client.ChunkedInput;
import org.glassfish.jersey.media.sse.EventListener;
import org.glassfish.jersey.media.sse.EventSource;
import org.glassfish.jersey.media.sse.EventSource.Builder;
import org.glassfish.jersey.media.sse.InboundEvent;
import org.glassfish.jersey.media.sse.SseFeature;

import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.Response;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.CompletionStage;
import java.util.function.BiConsumer;

public class HTTPClient extends InputAdapter implements Parameterizable, Runnable {

    private static final String STATUS_CONNECT_FAILED = "Connect Failed";
    private static final String STATUS_MESSAGE_FAILURE = "Message Failure";
    private static final String STATUS_CONNECTING = "Connecting";
    private static final String STATUS_DISCONNECTED = "Disconnected";
    private static final String STATUS_CONNECTED = "Connected";

    public static enum HTTPStreamType {
        HTTP_STREAM_TYPE_SSE("Server-Sent Events"),
        HTTP_STREAM_TYPE_CHUNKED("Chunked Response");

        private final String humanReadable;

        private HTTPStreamType(String s) {
            humanReadable = s;
        }

        @Override
        public String toString() {
            return humanReadable;
        }
    }

    public static final String FIELD_COMMAND = "command";
    public static final String FIELD_URL = "url";
    public static final String FIELD_HEADERS = "headers";
    public static final String FIELD_KEY = "key";
    public static final String FIELD_VALUE = "value";
    public static final String FIELD_ID = "id";
    public static final String FIELD_TYPE = "type";
    public static final String FIELD_COMMENT = "comment";
    public static final String FIELD_DATA = "data";

    public static final String STATUS_FIELD_ERROR = "Error";
    public static final String STATUS_FIELD_RESPONSE = "Response";
    public static final String STATUS_FIELD_URL = "URL";

    public static final String COMMAND_CONNECT = "connect";
    public static final String COMMAND_DISCONNECT = "disconnect";

    public static final int CONTROL_INPUT_PORT = 0;
    public static final int DATA_OUTPUT_PORT = 0;
    public static final int STATUS_OUTPUT_PORT = 1;
    private Logger logger;
    protected LogLevel logLevel = LogLevel.INFO;

    protected HTTPStreamType httpStreamType = HTTPStreamType.HTTP_STREAM_TYPE_SSE;
    public static final long serialVersionUID = 1567777700071L;

    private String url;
    private int connectTimeoutSeconds = 30;
    private int readTimeoutSeconds = 0;
    private int reconnectTimeoutSeconds = 3;
    private String displayName = "Server-Sent Events";
    private StatusPort statusPort;
    private String[][] httpHeaders = new String[0][0];
    private BlockingQueue<Tuple> commandQueue = new LinkedBlockingQueue<>();

    public HTTPClient() {
        super();
        logger = getLogger();
        setPortHints(1, 2);
        setDisplayName(displayName);
        setShortDisplayName(this.getClass().getSimpleName());
        setUrl("");

    }

    private Schema createControlHeaderSchema(String mainName) {
        return new Schema(Util.isEmpty(mainName) ? "" : mainName + "ControlHeaderSchema",
                          new Schema.Field(FIELD_KEY, CompleteDataType.forString()),
                          new Schema.Field(FIELD_VALUE, CompleteDataType.forString()));

    }

    private Schema createControlSchema(String mainName) {
        return new Schema(Util.isEmpty(mainName) ? "" : mainName + "ControlSchema",
                          new Schema.Field(FIELD_COMMAND, CompleteDataType.forString()),
                          new Schema.Field(FIELD_URL, CompleteDataType.forString()),
                          new Schema.Field(FIELD_HEADERS, CompleteDataType.forList(CompleteDataType.forTuple(createControlHeaderSchema(mainName)))));
    }

    public Schema[] getProposedInputSchemas(String mainName) {
        return new Schema[] { createControlHeaderSchema(mainName), createControlSchema(mainName) };
    }

    public void typecheck() throws TypecheckException {
        Set<String> optionalFields = new HashSet<>();
        optionalFields.add(FIELD_HEADERS);

        AdapterUtil.validateSchema(this, createControlSchema(""), getInputSchema(CONTROL_INPUT_PORT), true, CONTROL_INPUT_PORT, "control", optionalFields, false, null);

        switch (getHttpStreamType()) {
        case HTTP_STREAM_TYPE_CHUNKED:
            setOutputSchema(DATA_OUTPUT_PORT, new Schema("",
                                                         new Schema.Field(FIELD_DATA, CompleteDataType.forString()),
                                                         new Schema.Field(FIELD_URL, CompleteDataType.forString())));
            break;
        case HTTP_STREAM_TYPE_SSE:
            setOutputSchema(DATA_OUTPUT_PORT, new Schema("",
                                                         new Schema.Field(FIELD_ID, CompleteDataType.forString()),
                                                         new Schema.Field(FIELD_TYPE, CompleteDataType.forString()),
                                                         new Schema.Field(FIELD_DATA, CompleteDataType.forString()),
                                                         new Schema.Field(FIELD_COMMENT, CompleteDataType.forString()),                                                         
                                                         new Schema.Field(FIELD_URL, CompleteDataType.forString())));
            break;
        default:
            throw new TypecheckException("Invalid stream type");
        }

        statusPort = new StatusPort(STATUS_OUTPUT_PORT, this, true, logger);
        setOutputSchema(STATUS_OUTPUT_PORT, statusPort.getSchema());
    }

    public void init() throws StreamBaseException {
        super.init();
        logger = AdapterUtil.getLogger(this, getLogLevel());
        statusPort.setSchema(getOutputSchema(STATUS_OUTPUT_PORT));
        statusPort.setLogger(logger);
        registerRunnable("Server-Sent Events", this, true);
    }

    @Override
    public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {
        commandQueue.add(tuple);
    }

    @Override
    public void resumed() throws Exception {
        super.resumed();
        if (!Util.isEmpty(url)) {
            Tuple tuple = getInputSchema(CONTROL_INPUT_PORT).createTuple();
            tuple.setString(FIELD_COMMAND, COMMAND_DISCONNECT);
            tuple.setString(FIELD_URL, getUrl());
            List<Tuple> headers = new ArrayList<>();
            for (int x = 0; x < httpHeaders.length; x++) {
                Tuple headerTuple = tuple.getSchema().getField(FIELD_HEADERS).getElementType().getSchema().createTuple();
                headerTuple.setString(FIELD_KEY, httpHeaders[x][0]);
                headerTuple.setString(FIELD_VALUE, httpHeaders[x][1]);
                headers.add(headerTuple);
            }
            commandQueue.add(tuple);
        }
    }
    
    private Map<String, String> combineAllHeaders(Tuple tuple) throws StreamBaseException {
        Map<String, String> headers = new HashMap<>();
        for (String[] defaultHeader : getHttpHeaders()) {
            String key = defaultHeader[0];
            String value = defaultHeader[1];
            if (!Util.isEmpty(key)) {
                headers.put(key, value);
            }
        }
        List<?> tupleHeaders = tuple.isNull(FIELD_HEADERS) ? null : tuple.getList(FIELD_HEADERS);
        if (tupleHeaders != null) {
            for (Object headerObj : tupleHeaders) {
                if (headerObj != null) {
                    Tuple headerTuple = (Tuple) headerObj;
                    String key = headerTuple.isNull(FIELD_KEY) ? "" : headerTuple.getString(FIELD_KEY);
                    String value = headerTuple.isNull(FIELD_VALUE) ? "" : headerTuple.getString(FIELD_VALUE);
                    if (!Util.isEmpty(key)) {
                        headers.put(key, value);
                    }
                }
            }
        }
        return headers;
    }

    @Override
    public void run() {
        while (shouldRun()) {
            try {
                Tuple tuple = commandQueue.poll(100, TimeUnit.MILLISECONDS);
                if (tuple != null) {
                    try {
                        String url = tuple.isNull(FIELD_URL) ? getUrl() : tuple.getString(FIELD_URL);
                        url = Util.isEmpty(url) ? getUrl() : url;
                        if (!Util.isEmpty(url)) {
                            String command = tuple.isNull(FIELD_COMMAND) ? "" : tuple.getString(FIELD_COMMAND);

                            if (command.equalsIgnoreCase(COMMAND_DISCONNECT)) {
                                switch (httpStreamType) {
                                case HTTP_STREAM_TYPE_SSE:
                                default:
                                    stopSSE(url);
                                    break;
                                case HTTP_STREAM_TYPE_CHUNKED:
                                    stopChunked(url);
                                    break;
                                }
                            } else if (command.equalsIgnoreCase(COMMAND_CONNECT) || Util.isEmpty(command)) {
                                switch (httpStreamType) {
                                case HTTP_STREAM_TYPE_SSE:
                                default:
                                    startSSE(tuple, url);
                                    break;
                                case HTTP_STREAM_TYPE_CHUNKED:
                                    startChunked(tuple, url);
                                    break;
                                }
                            } else {
                                statusPort.onStatus("Invalid Command", "Valid Commands", "connect/disconnect");
                            }
                        }
                    } catch (StreamBaseException e) {
                    }
                }
            } catch (InterruptedException e) {

            }
        }
        for (Entry<String, CompletionStage<Response>> entry : chunkedResponses.entrySet()) {
            entry.getValue().toCompletableFuture().cancel(true);
        }
    }

    Map<String, EventSource> eventSources = new HashMap<>();
    private void stopSSE(String url) {
        EventSource eventSource = eventSources.remove(url);
        if (eventSource != null) {
            eventSource.close();
            getStatusPort().onStatus(STATUS_DISCONNECTED, HTTPClient.FIELD_URL, url);
        }
    }

    private void startSSE(Tuple tuple, String url) throws StreamBaseException {
        getStatusPort().onStatus(STATUS_CONNECTING, HTTPClient.FIELD_URL, url);
        Map<String, String> headers = combineAllHeaders(tuple);
        Client client = ClientBuilder.newBuilder().readTimeout(getReadTimeoutSeconds(), TimeUnit.SECONDS).connectTimeout(getConnectTimeoutSeconds(), TimeUnit.SECONDS).register(SseFeature.class).register(new AddHeaderOnRequestFilter(headers)).build();      
        WebTarget target = client.target(url);              
        Builder source = EventSource.target(target);        
        source.reconnectingEvery(getReconnectTimeoutSeconds(), TimeUnit.SECONDS);        
        EventSource eventSource = source.build();        
        eventSource.register(new EventListener() {
            
            @Override
            public void onEvent(InboundEvent event) {                
                if (event != null && !event.isEmpty()) {
                    Tuple tuple = getOutputSchema(HTTPClient.DATA_OUTPUT_PORT).createTuple();
                    try {
                        tuple.setString(HTTPClient.FIELD_ID, event.getId());
                        tuple.setString(HTTPClient.FIELD_DATA, event.readData());
                        tuple.setString(HTTPClient.FIELD_COMMENT, event.getComment());                        
                        tuple.setString(HTTPClient.FIELD_TYPE, event.getName());
                        tuple.setString(HTTPClient.FIELD_URL, url);
                        sendOutput(0, tuple);
                    } catch (StreamBaseException e) {
                        getStatusPort().onStatus(LogLevel.ERROR, e, STATUS_MESSAGE_FAILURE, HTTPClient.FIELD_URL, url, STATUS_FIELD_ERROR, e.getMessage());
                    }
                }
            }
        });
        eventSource.open();
        getStatusPort().onStatus(STATUS_CONNECTED, HTTPClient.FIELD_URL, url);
        eventSources.put(url, eventSource);
    }

    Map<String, CompletionStage<Response>> chunkedResponses = new HashMap<>();
    private void stopChunked(String url) {
        CompletionStage<Response> response = chunkedResponses.remove(url);
        if (response != null) {
            response.toCompletableFuture().cancel(true);
            getStatusPort().onStatus(STATUS_DISCONNECTED, HTTPClient.FIELD_URL, url);
        }
    }
        
    private void startChunked(final Tuple tuple, final String url) throws StreamBaseException {
        getStatusPort().onStatus(STATUS_CONNECTING, HTTPClient.FIELD_URL, url);
        Client client = ClientBuilder.newBuilder().readTimeout(getReadTimeoutSeconds(), TimeUnit.SECONDS).connectTimeout(getConnectTimeoutSeconds(), TimeUnit.SECONDS).build();
        WebTarget target = client.target(url);
        Map<String, String> headers = combineAllHeaders(tuple);
        for (Entry<String, String> entry : headers.entrySet()) {
            target.request().header(entry.getKey(), entry.getValue());
        }

        CompletionStage<Response> response = target.request().rx().get();
        response.whenComplete(new BiConsumer<Response, Throwable>() {
            @Override
            public void accept(Response response, Throwable ex) {
                if (ex != null) {
                    getStatusPort().onStatus(LogLevel.ERROR, new Exception(ex), STATUS_CONNECT_FAILED, HTTPClient.FIELD_URL, url, STATUS_FIELD_ERROR, ex.getMessage());
                    return;
                }
                getStatusPort().onStatus(STATUS_CONNECTED, HTTPClient.FIELD_URL, url);
                final ChunkedInput<String> chunkedInput = response.readEntity(new GenericType<ChunkedInput<String>>() {
                });
                String chunk;
                Integer lastLength = null;
                StringBuilder sb = new StringBuilder();
                while ((chunk = chunkedInput.read()) != null) {
                    if (lastLength == null) {
                    	sb.setLength(0);
                    	try {
                    		lastLength = Integer.parseInt(chunk, 16);                                    
                            continue;
                    	} catch (NumberFormatException e) {
                    		lastLength = chunk.length();
                    	}                                
                    }
                    sb.append(chunk);
                    lastLength -= chunk.length();
                    if (lastLength <= 0) {
	                    Tuple tuple = getOutputSchema(HTTPClient.DATA_OUTPUT_PORT).createTuple();
	                    try {
	                        tuple.setString(HTTPClient.FIELD_DATA, chunk);
	                        tuple.setString(HTTPClient.FIELD_URL, url);
	                        sendOutput(0, tuple);
	                    } catch (StreamBaseException e) {
	                    	getStatusPort().onStatus(LogLevel.ERROR, e, STATUS_MESSAGE_FAILURE, HTTPClient.FIELD_URL, url, STATUS_FIELD_ERROR, e.getMessage());
                        }
                        lastLength = null;
                    }
                }
                chunkedResponses.remove(url);
                getStatusPort().onStatus(STATUS_DISCONNECTED, HTTPClient.FIELD_URL, url);
            }
        });        
        chunkedResponses.put(url, response);
    }

    public void setUrl(String url) {
        this.url = url;
    }

    public String getUrl() {
        return this.url;
    }

    public void setConnectTimeoutSeconds(int connectTimeoutSeconds) {
        this.connectTimeoutSeconds = connectTimeoutSeconds;
    }

    public int getConnectTimeoutSeconds() {
        return this.connectTimeoutSeconds;
    }

    public void setReadTimeoutSeconds(int readTimeoutSeconds) {
        this.readTimeoutSeconds = readTimeoutSeconds;
    }

    public int getReadTimeoutSeconds() {
        return this.readTimeoutSeconds;
    }

    public void setReconnectTimeoutSeconds(int reconnectTimeoutSeconds) {
        this.reconnectTimeoutSeconds = reconnectTimeoutSeconds;
    }

    public int getReconnectTimeoutSeconds() {
        return this.reconnectTimeoutSeconds;
    }

    public void setHttpHeaders(String[][] httpHeaders) {
        this.httpHeaders = httpHeaders;
    }

    public String[][] getHttpHeaders() {
        return this.httpHeaders;
    }

    public HTTPStreamType getHttpStreamType() {
        return httpStreamType;
    }

    public void setHttpStreamType(HTTPStreamType httpStreamType) {
        this.httpStreamType = httpStreamType;
    }

    public LogLevel getLogLevel() {
        return logLevel;
    }

    public void setLogLevel(LogLevel logLevel) {
        this.logLevel = logLevel;
    }

    public StatusPort getStatusPort() {
        return statusPort;
    }
}
